@@ -105,7 +105,7 @@ void QQuick3DObject::update()
 }
 
 /*!
-    \qmlproperty Object3D QtQuick::Object3D::parent
+    \qmlproperty Object3D QtQuick3D::Object3D::parent
     This property holds the parent of the Object3D in a 3D scene.
 
     \note An Object3D's parent may not necessarily be the same as its object
@@ -1101,7 +1101,7 @@ void QV4::Heap::QSSGItemWrapper::markObjects(QV4::Heap::Base *that, QV4::MarkSta
     QObjectWrapper::markObjects(that, markStack);
 }
 
-quint64 QQuick3DObjectPrivate::_q_createJSWrapper(QV4::ExecutionEngine *engine)
+quint64 QQuick3DObjectPrivate::_q_createJSWrapper(QQmlV4ExecutionEnginePtr engine)
 {
     return (engine->memoryManager->allocate<QSSGItemWrapper>(q_func()))->asReturnedValue();
 }
