@@ -4,11 +4,13 @@
 
 #include "graphobjects/qssgrendergraphobject_p.h"
 #include "qssgrendershadercache_p.h"
+#include "qssgrendercontextcore.h"
 
 #include <QtQuick3DUtils/private/qssgutils_p.h>
 #include <QtQuick3DUtils/private/qquick3dprofiler_p.h>
 
 #include <QtQuick3DRuntimeRender/private/qssgruntimerenderlogging_p.h>
+#include <qtquick3d_tracepoints_p.h>
 
 #include <QCoreApplication>
 #include <QStandardPaths>
@@ -19,17 +21,19 @@
 #include <QtGui/qsurfaceformat.h>
 #if QT_CONFIG(opengl)
 # include <QtGui/qopenglcontext.h>
-# include <QtGui/private/qrhigles2_p.h>
 #endif
 
 #ifdef QT_QUICK3D_HAS_RUNTIME_SHADERS
-#include <QtShaderTools/private/qshaderbaker_p.h>
+#include <rhi/qshaderbaker.h>
 #endif
 
 #include <QtCore/qmutex.h>
 
 QT_BEGIN_NAMESPACE
 
+Q_TRACE_POINT(qtquick3d, QSSG_loadShader_entry)
+Q_TRACE_POINT(qtquick3d, QSSG_loadShader_exit)
+
 static QtQuick3DEditorHelpers::ShaderBaker::StatusCallback s_statusCallback = nullptr;
 Q_GLOBAL_STATIC(QMutex, s_statusMutex);
 
@@ -71,7 +75,8 @@ static constexpr DefineEntry DefineTable[] {
     { "QSSG_ENABLE_OPAQUE_DEPTH_PRE_PASS", QSSGShaderFeatures::Feature::OpaqueDepthPrePass },
     { "QSSG_ENABLE_REFLECTION_PROBE", QSSGShaderFeatures::Feature::ReflectionProbe },
     { "QSSG_REDUCE_MAX_NUM_LIGHTS", QSSGShaderFeatures::Feature::ReduceMaxNumLights },
-    { "QSSG_ENABLE_LIGHTMAP", QSSGShaderFeatures::Feature::Lightmap }
+    { "QSSG_ENABLE_LIGHTMAP", QSSGShaderFeatures::Feature::Lightmap },
+    { "QSSG_DISABLE_MULTIVIEW", QSSGShaderFeatures::Feature::DisableMultiView }
 };
 
 static_assert(std::size(DefineTable) == QSSGShaderFeatures::Count, "Missing feature define?");
@@ -95,6 +100,9 @@ static void initBakerForNonPersistentUse(QShaderBaker *baker, QRhi *rhi)
     case QRhi::D3D11:
         outputs.append({ QShader::HlslShader, QShaderVersion(50) }); // Shader Model 5.0
         break;
+    case QRhi::D3D12:
+        outputs.append({ QShader::HlslShader, QShaderVersion(61) }); // Shader Model 6.1 (includes multiview support)
+        break;
     case QRhi::Metal:
         outputs.append({ QShader::MslShader, QShaderVersion(12) }); // Metal 1.2
         break;
@@ -156,6 +164,7 @@ static void initBakerForPersistentUse(QShaderBaker *baker, QRhi *)
     outputs.reserve(8);
     outputs.append({ QShader::SpirvShader, QShaderVersion(100) });
     outputs.append({ QShader::HlslShader, QShaderVersion(50) }); // Shader Model 5.0
+    outputs.append({ QShader::HlslShader, QShaderVersion(61) }); // Shader Model 6.1 (for multiview on d3d12)
     outputs.append({ QShader::MslShader, QShaderVersion(12) }); // Metal 1.2
     outputs.append({ QShader::GlslShader, QShaderVersion(330) }); // OpenGL 3.3+
     outputs.append({ QShader::GlslShader, QShaderVersion(140) }); // OpenGL 3.1+
@@ -191,6 +200,18 @@ static void initBakerForPersistentUse(QShaderBaker *, QRhi *)
 }
 #endif // QT_QUICK3D_HAS_RUNTIME_SHADERS
 
+static bool s_autoDiskCacheEnabled = true;
+
+static bool isAutoDiskCacheEnabled()
+{
+    // these three mirror QOpenGLShaderProgram/QQuickGraphicsConfiguration/QSGRhiSupport
+    static const bool diskCacheDisabled = qEnvironmentVariableIntValue("QT_DISABLE_SHADER_DISK_CACHE")
+                                          || qEnvironmentVariableIntValue("QSG_RHI_DISABLE_DISK_CACHE");
+    const bool attrDiskCacheDisabled = (qApp ? qApp->testAttribute(Qt::AA_DisableShaderDiskCache) : false);
+    return (!diskCacheDisabled && !attrDiskCacheDisabled && s_autoDiskCacheEnabled);
+
+}
+
 static inline bool ensureWritableDir(const QString &name)
 {
     QDir::root().mkpath(name);
@@ -227,22 +248,18 @@ static inline QString persistentQsbcFileName()
     return QString();
 }
 
-QSSGShaderCache::QSSGShaderCache(const QSSGRef<QSSGRhiContext> &ctx,
+QSSGShaderCache::QSSGShaderCache(QSSGRhiContext &ctx,
                                  const InitBakerFunc initBakeFn)
     : m_rhiContext(ctx),
-      m_initBaker(initBakeFn)
+      m_initBaker(initBakeFn),
+      m_builtInShaders(*this)
 {
-    // Mirrors QOpenGLShaderProgram/QQuickGraphicsConfiguration/QSGRhiSupport.
-    m_autoDiskCacheEnabled = !QCoreApplication::instance()->testAttribute(Qt::AA_DisableShaderDiskCache)
-            && !qEnvironmentVariableIntValue("QT_DISABLE_SHADER_DISK_CACHE")
-            && !qEnvironmentVariableIntValue("QSG_RHI_DISABLE_DISK_CACHE");
-
-    if (m_autoDiskCacheEnabled) {
-        const bool shaderDebug = !QSSGRhiContext::editorMode() && QSSGRhiContext::shaderDebuggingEnabled();
+    if (isAutoDiskCacheEnabled()) {
+        const bool shaderDebug = !QSSGRhiContextPrivate::editorMode() && QSSGRhiContextPrivate::shaderDebuggingEnabled();
         m_persistentShaderStorageFileName = persistentQsbcFileName();
         if (!m_persistentShaderStorageFileName.isEmpty()) {
             const bool skipCacheFile = qEnvironmentVariableIntValue("QT_QUICK3D_NO_SHADER_CACHE_LOAD");
-            if (!skipCacheFile && QFileInfo(m_persistentShaderStorageFileName).exists()) {
+            if (!skipCacheFile && QFileInfo::exists(m_persistentShaderStorageFileName)) {
                 if (shaderDebug)
                     qDebug("Attempting to seed material shader cache from %s", qPrintable(m_persistentShaderStorageFileName));
                 if (m_persistentShaderBakingCache.load(m_persistentShaderStorageFileName)) {
@@ -276,13 +293,15 @@ QSSGShaderCache::~QSSGShaderCache()
 
 void QSSGShaderCache::releaseCachedResources()
 {
+    m_builtInShaders.releaseCachedResources();
+
     m_rhiShaders.clear();
 
     // m_persistentShaderBakingCache is not cleared, that is intentional,
     // otherwise we would permanently lose what got loaded at startup.
 }
 
-QSSGRef<QSSGRhiShaderPipeline> QSSGShaderCache::tryGetRhiShaderPipeline(const QByteArray &inKey,
+QSSGRhiShaderPipelinePtr QSSGShaderCache::tryGetRhiShaderPipeline(const QByteArray &inKey,
                                                                         const QSSGShaderFeatures &inFeatures)
 {
     QSSGShaderCacheKey cacheKey(inKey);
@@ -298,7 +317,8 @@ QSSGRef<QSSGRhiShaderPipeline> QSSGShaderCache::tryGetRhiShaderPipeline(const QB
 void QSSGShaderCache::addShaderPreprocessor(QByteArray &str,
                                             const QByteArray &inKey,
                                             ShaderType shaderType,
-                                            const QSSGShaderFeatures &inFeatures)
+                                            const QSSGShaderFeatures &inFeatures,
+                                            int viewCount)
 {
     m_insertStr.clear();
 
@@ -312,6 +332,11 @@ void QSSGShaderCache::addShaderPreprocessor(QByteArray &str,
 
     m_insertStr += "#define texture2D texture\n";
 
+    // match Qt Quick and QSGMaterial(Shader)
+    m_insertStr += "#define QSHADER_VIEW_COUNT ";
+    m_insertStr += QByteArray::number(viewCount);
+    m_insertStr += "\n";
+
     str.insert(0, m_insertStr);
     QString::size_type insertPos = int(m_insertStr.size());
 
@@ -345,11 +370,13 @@ QByteArray QSSGShaderCache::shaderCollectionFile()
     return QByteArrayLiteral("qtappshaders.qsbc");
 }
 
-QSSGRef<QSSGRhiShaderPipeline> QSSGShaderCache::compileForRhi(const QByteArray &inKey, const QByteArray &inVert, const QByteArray &inFrag,
-                                                              const QSSGShaderFeatures &inFeatures, QSSGRhiShaderPipeline::StageFlags stageFlags)
+QSSGRhiShaderPipelinePtr QSSGShaderCache::compileForRhi(const QByteArray &inKey, const QByteArray &inVert, const QByteArray &inFrag,
+                                                        const QSSGShaderFeatures &inFeatures, QSSGRhiShaderPipeline::StageFlags stageFlags,
+                                                        int viewCount,
+                                                        bool perTargetCompilation)
 {
 #ifdef QT_QUICK3D_HAS_RUNTIME_SHADERS
-    const QSSGRef<QSSGRhiShaderPipeline> &rhiShaders = tryGetRhiShaderPipeline(inKey, inFeatures);
+    const QSSGRhiShaderPipelinePtr &rhiShaders = tryGetRhiShaderPipeline(inKey, inFeatures);
     if (rhiShaders)
         return rhiShaders;
 
@@ -361,22 +388,31 @@ QSSGRef<QSSGRhiShaderPipeline> QSSGShaderCache::compileForRhi(const QByteArray &
     QByteArray fragmentCode = inFrag;
 
     if (!vertexCode.isEmpty())
-        addShaderPreprocessor(vertexCode, inKey, ShaderType::Vertex, inFeatures);
+        addShaderPreprocessor(vertexCode, inKey, ShaderType::Vertex, inFeatures, viewCount);
 
     if (!fragmentCode.isEmpty())
-        addShaderPreprocessor(fragmentCode, inKey, ShaderType::Fragment, inFeatures);
+        addShaderPreprocessor(fragmentCode, inKey, ShaderType::Fragment, inFeatures, viewCount);
 
     // lo and behold the final shader strings are ready
 
-    QSSGRef<QSSGRhiShaderPipeline> shaders;
+    QSSGRhiShaderPipelinePtr shaders;
     QString vertErr, fragErr;
 
     QShaderBaker baker;
-    m_initBaker(&baker, m_rhiContext->rhi());
+    m_initBaker(&baker, m_rhiContext.rhi());
 
-    const bool editorMode = QSSGRhiContext::editorMode();
+    // If requested, per-target compilation allows doing things like #if
+    // QSHADER_HLSL in the shader code, at the expense of spending more time in
+    // bake())
+    baker.setPerTargetCompilation(perTargetCompilation);
+
+    // This is in the shader key, but cannot query that here anymore now that it's serialized.
+    // So we get it as a dedicated argument.
+    baker.setMultiViewCount(viewCount);
+
+    const bool editorMode = QSSGRhiContextPrivate::editorMode();
     // Shader debug is disabled in editor mode
-    const bool shaderDebug = !editorMode && QSSGRhiContext::shaderDebuggingEnabled();
+    const bool shaderDebug = !editorMode && QSSGRhiContextPrivate::shaderDebuggingEnabled();
 
    static auto dumpShader = [](QShader::Stage stage, const QByteArray &code) {
        switch (stage) {
@@ -409,9 +445,9 @@ QSSGRef<QSSGRhiShaderPipeline> QSSGShaderCache::compileForRhi(const QByteArray &
     if (!vertShaderValid) {
         vertErr = baker.errorMessage();
         if (!editorMode) {
-            qWarning("Failed to compile vertex shader:\n");
+            qWarning("Failed to compile vertex shader: %s\n", qPrintable(vertErr));
             if (!shaderDebug)
-                qWarning() << inKey << '\n' << vertErr;
+                qWarning() << inKey << '\n';
         }
     }
 
@@ -427,9 +463,9 @@ QSSGRef<QSSGRhiShaderPipeline> QSSGShaderCache::compileForRhi(const QByteArray &
     if (!fragShaderValid) {
         fragErr = baker.errorMessage();
         if (!editorMode) {
-            qWarning("Failed to compile fragment shader \n");
+            qWarning("Failed to compile fragment shader: %s\n", qPrintable(fragErr));
             if (!shaderDebug)
-                qWarning() << inKey << '\n' << fragErr;
+                qWarning() << inKey << '\n';
         }
     }
 
@@ -440,7 +476,7 @@ QSSGRef<QSSGRhiShaderPipeline> QSSGShaderCache::compileForRhi(const QByteArray &
     }
 
     if (vertShaderValid && fragShaderValid) {
-        shaders = new QSSGRhiShaderPipeline(*m_rhiContext.data());
+        shaders = std::make_shared<QSSGRhiShaderPipeline>(m_rhiContext);
         shaders->addStage(QRhiShaderStage(QRhiShaderStage::Vertex, vertexShader), stageFlags);
         shaders->addStage(QRhiShaderStage(QRhiShaderStage::Fragment, fragmentShader), stageFlags);
         if (shaderDebug)
@@ -456,7 +492,7 @@ QSSGRef<QSSGRhiShaderPipeline> QSSGShaderCache::compileForRhi(const QByteArray &
         s_statusCallback(inKey, fragStatus, fragErr, QShader::FragmentStage);
     }
 
-    QSSGRef<QSSGRhiShaderPipeline> result = m_rhiShaders.insert(tempKey, shaders).value();
+    auto result = m_rhiShaders.insert(tempKey, shaders).value();
     if (result && result->vertexStage() && result->fragmentStage()) {
         QQsbCollection::EntryDesc entryDesc = {
             inKey,
@@ -480,11 +516,11 @@ QSSGRef<QSSGRhiShaderPipeline> QSSGShaderCache::compileForRhi(const QByteArray &
 #endif
 }
 
-QSSGRef<QSSGRhiShaderPipeline> QSSGShaderCache::newPipelineFromPregenerated(const QByteArray &inKey,
-                                                                            const QSSGShaderFeatures &inFeatures,
-                                                                            QQsbCollection::Entry entry,
-                                                                            const QSSGRenderGraphObject &obj,
-                                                                            QSSGRhiShaderPipeline::StageFlags stageFlags)
+QSSGRhiShaderPipelinePtr QSSGShaderCache::newPipelineFromPregenerated(const QByteArray &inKey,
+                                                                      const QSSGShaderFeatures &inFeatures,
+                                                                      QQsbCollection::Entry entry,
+                                                                      const QSSGRenderGraphObject &obj,
+                                                                      QSSGRhiShaderPipeline::StageFlags stageFlags)
 {
     // No lookup in m_rhiShaders. It is up to the caller to do that, if they
     // want to. We will insert into it at the end, but there is intentionally
@@ -493,14 +529,15 @@ QSSGRef<QSSGRhiShaderPipeline> QSSGShaderCache::newPipelineFromPregenerated(cons
     // full-blown generator). That is important for some clients (effect
     // system) so returning an existing QSSGRhiShaderPipeline is _wrong_.
 
-    const bool shaderDebug = !QSSGRhiContext::editorMode() && QSSGRhiContext::shaderDebuggingEnabled();
+    const bool shaderDebug = !QSSGRhiContextPrivate::editorMode() && QSSGRhiContextPrivate::shaderDebuggingEnabled();
     if (shaderDebug)
         qDebug("Loading pregenerated rhi shader(s)");
 
+    Q_TRACE_SCOPE(QSSG_loadShader);
     Q_QUICK3D_PROFILE_START(QQuick3DProfiler::Quick3DLoadShader);
 
     // Note that we are required to return a non-null (but empty) shader set even if loading fails.
-    QSSGRef<QSSGRhiShaderPipeline> shaders(new QSSGRhiShaderPipeline(*m_rhiContext.data()));
+    QSSGRhiShaderPipelinePtr shaders(new QSSGRhiShaderPipeline(m_rhiContext));
 
     const QString collectionFile = QString::fromLatin1(resourceFolder() + shaderCollectionFile());
 
@@ -518,7 +555,11 @@ QSSGRef<QSSGRhiShaderPipeline> QSSGShaderCache::newPipelineFromPregenerated(cons
             qDebug("Loading of vertex and fragment stages succeeded");
     }
 
+#if !QT_CONFIG(qml_debug)
+    Q_UNUSED(obj);
+#else
     Q_QUICK3D_PROFILE_END_WITH_ID(QQuick3DProfiler::Quick3DLoadShader, 0, obj.profilingId);
+#endif
 
     QSSGShaderCacheKey cacheKey(inKey);
     cacheKey.m_features = inFeatures;
@@ -529,10 +570,10 @@ QSSGRef<QSSGRhiShaderPipeline> QSSGShaderCache::newPipelineFromPregenerated(cons
     return inserted.value();
 }
 
-QSSGRef<QSSGRhiShaderPipeline> QSSGShaderCache::tryNewPipelineFromPersistentCache(const QByteArray &qsbcKey,
-                                                                                  const QByteArray &inKey,
-                                                                                  const QSSGShaderFeatures &inFeatures,
-                                                                                  QSSGRhiShaderPipeline::StageFlags stageFlags)
+QSSGRhiShaderPipelinePtr QSSGShaderCache::tryNewPipelineFromPersistentCache(const QByteArray &qsbcKey,
+                                                                            const QByteArray &inKey,
+                                                                            const QSSGShaderFeatures &inFeatures,
+                                                                            QSSGRhiShaderPipeline::StageFlags stageFlags)
 {
     // No lookup in m_rhiShaders. it is up to the caller to do that, if they
     // want to. We will insert into it at the end, but there is intentionally
@@ -549,11 +590,11 @@ QSSGRef<QSSGRhiShaderPipeline> QSSGShaderCache::tryNewPipelineFromPersistentCach
         return {};
 
     if (entryDesc.vertShader.isValid() && entryDesc.fragShader.isValid()) {
-        const bool shaderDebug = !QSSGRhiContext::editorMode() && QSSGRhiContext::shaderDebuggingEnabled();
+        const bool shaderDebug = !QSSGRhiContextPrivate::editorMode() && QSSGRhiContextPrivate::shaderDebuggingEnabled();
         if (shaderDebug)
             qDebug("Loading rhi shaders from disk cache for %s (%s)", qsbcKey.constData(), inKey.constData());
 
-        QSSGRef<QSSGRhiShaderPipeline> shaders(new QSSGRhiShaderPipeline(*m_rhiContext.data()));
+        QSSGRhiShaderPipelinePtr shaders(new QSSGRhiShaderPipeline(m_rhiContext));
         shaders->addStage(QRhiShaderStage(QRhiShaderStage::Vertex, entryDesc.vertShader), stageFlags);
         shaders->addStage(QRhiShaderStage(QRhiShaderStage::Fragment, entryDesc.fragShader), stageFlags);
         QSSGShaderCacheKey cacheKey(inKey);
@@ -565,27 +606,31 @@ QSSGRef<QSSGRhiShaderPipeline> QSSGShaderCache::tryNewPipelineFromPersistentCach
     return {};
 }
 
-QSSGRef<QSSGRhiShaderPipeline> QSSGShaderCache::loadBuiltinForRhi(const QByteArray &inKey)
+QSSGRhiShaderPipelinePtr QSSGShaderCache::loadBuiltinUncached(const QByteArray &inKey, int viewCount)
 {
-    const QSSGRef<QSSGRhiShaderPipeline> &rhiShaders = tryGetRhiShaderPipeline(inKey, QSSGShaderFeatures());
-    if (rhiShaders)
-        return rhiShaders;
-
-    const bool shaderDebug = !QSSGRhiContext::editorMode() && QSSGRhiContext::shaderDebuggingEnabled();
+    const bool shaderDebug = !QSSGRhiContextPrivate::editorMode() && QSSGRhiContextPrivate::shaderDebuggingEnabled();
     if (shaderDebug)
-        qDebug("Loading builtin rhi shader: %s", inKey.constData());
+        qDebug("Loading builtin rhi shader: %s (view count: %d)", inKey.constData(), viewCount);
 
+    Q_TRACE_SCOPE(QSSG_loadShader);
     Q_QUICK3D_PROFILE_START(QQuick3DProfiler::Quick3DLoadShader);
 
     // Note that we are required to return a non-null (but empty) shader set even if loading fails.
-    QSSGRef<QSSGRhiShaderPipeline> shaders(new QSSGRhiShaderPipeline(*m_rhiContext.data()));
+    QSSGRhiShaderPipelinePtr shaders(new QSSGRhiShaderPipeline(m_rhiContext));
 
     // inShaderName is a prefix of a .qsb file, so "abc" means we should
     // look for abc.vert.qsb and abc.frag.qsb.
 
     const QString prefix = QString::fromUtf8(resourceFolder() + inKey);
-    const QString vertexFileName = prefix + QLatin1String(".vert.qsb");
-    const QString fragmentFileName = prefix + QLatin1String(".frag.qsb");
+    QString vertexFileName = prefix + QLatin1String(".vert.qsb");
+    QString fragmentFileName = prefix + QLatin1String(".frag.qsb");
+
+    // This must match QSGMaterial(Shader) in Qt Quick, in particular the
+    // QSGMaterialShader::setShaderFileName() overload taking a viewCount.
+    if (viewCount == 2) {
+        vertexFileName += QLatin1String(".mv2qsb");
+        fragmentFileName += QLatin1String(".mv2qsb");
+    }
 
     QShader vertexShader;
     QShader fragmentShader;
@@ -617,12 +662,7 @@ QSSGRef<QSSGRhiShaderPipeline> QSSGShaderCache::loadBuiltinForRhi(const QByteArr
 
     Q_QUICK3D_PROFILE_END_WITH_STRING(QQuick3DProfiler::Quick3DLoadShader, 0, inKey);
 
-    QSSGShaderCacheKey cacheKey(inKey);
-    cacheKey.m_features = QSSGShaderFeatures();
-    cacheKey.updateHashCode();
-
-    const auto inserted = m_rhiShaders.insert(cacheKey, shaders);
-    return inserted.value();
+    return shaders;
 }
 
 namespace QtQuick3DEditorHelpers {
@@ -631,6 +671,17 @@ void ShaderBaker::setStatusCallback(StatusCallback cb)
     QMutexLocker locker(&*s_statusMutex);
     s_statusCallback = cb;
 }
+
+void ShaderCache::setAutomaticDiskCache(bool enable)
+{
+    s_autoDiskCacheEnabled = enable;
+}
+
+bool ShaderCache::isAutomaticDiskCacheEnabled()
+{
+    return ::isAutoDiskCacheEnabled();
+}
+
 }
 
 QT_END_NAMESPACE
