@@ -5,7 +5,8 @@
 /* clang-format off */
 
 #include <QtQuick3DRuntimeRender/private/qssgshadermaterialadapter_p.h>
-#include <QtQuick3DRuntimeRender/private/qssgrendercontextcore_p.h>
+#include "qssgrendercontextcore.h"
+#include <QtQuick3DRuntimeRender/private/qssgrhicustommaterialsystem_p.h>
 
 QT_BEGIN_NAMESPACE
 
@@ -40,20 +41,21 @@ bool QSSGShaderMaterialAdapter::hasCustomShaderSnippet(QSSGShaderCache::ShaderTy
 }
 
 QByteArray QSSGShaderMaterialAdapter::customShaderSnippet(QSSGShaderCache::ShaderType,
-                                                          const QSSGRef<QSSGShaderLibraryManager> &)
+                                                          QSSGShaderLibraryManager &,
+                                                          bool)
 {
     return QByteArray();
 }
 
 bool QSSGShaderMaterialAdapter::hasCustomShaderFunction(QSSGShaderCache::ShaderType,
                                                         const QByteArray &,
-                                                        const QSSGRef<QSSGShaderLibraryManager> &)
+                                                        QSSGShaderLibraryManager &)
 {
     return false;
 }
 
 void QSSGShaderMaterialAdapter::setCustomPropertyUniforms(char *,
-                                                          QSSGRef<QSSGRhiShaderPipeline> &,
+                                                          QSSGRhiShaderPipeline &,
                                                           const QSSGRenderContextInterface &)
 {
 }
@@ -95,6 +97,31 @@ bool QSSGShaderDefaultMaterialAdapter::isVertexColorsEnabled()
     return m_material.isVertexColorsEnabled();
 }
 
+bool QSSGShaderDefaultMaterialAdapter::isVertexColorsMaskEnabled()
+{
+    return m_material.isVertexColorsMaskEnabled();
+}
+
+bool QSSGShaderDefaultMaterialAdapter::isInvertOpacityMapValue()
+{
+    return m_material.isInvertOpacityMapValue();
+}
+
+bool QSSGShaderDefaultMaterialAdapter::isBaseColorSingleChannelEnabled()
+{
+    return m_material.isBaseColorSingleChannelEnabled();
+}
+
+bool QSSGShaderDefaultMaterialAdapter::isSpecularAmountSingleChannelEnabled()
+{
+    return m_material.isSpecularAmountSingleChannelEnabled();
+}
+
+bool QSSGShaderDefaultMaterialAdapter::isEmissiveSingleChannelEnabled()
+{
+    return m_material.isEmissiveSingleChannelEnabled();
+}
+
 bool QSSGShaderDefaultMaterialAdapter::isClearcoatEnabled()
 {
     return m_material.isClearcoatEnabled();
@@ -115,6 +142,11 @@ bool QSSGShaderDefaultMaterialAdapter::usesCustomSkinning()
     return false;
 }
 
+bool QSSGShaderDefaultMaterialAdapter::usesCustomMorphing()
+{
+    return false;
+}
+
 QSSGRenderDefaultMaterial::MaterialSpecularModel QSSGShaderDefaultMaterialAdapter::specularModel()
 {
     return m_material.specularModel;
@@ -125,6 +157,26 @@ QSSGRenderDefaultMaterial::MaterialAlphaMode QSSGShaderDefaultMaterialAdapter::a
     return m_material.alphaMode;
 }
 
+QSSGRenderDefaultMaterial::VertexColorMaskFlags QSSGShaderDefaultMaterialAdapter::vertexColorRedMask()
+{
+    return m_material.vertexColorRedMask;
+}
+
+QSSGRenderDefaultMaterial::VertexColorMaskFlags QSSGShaderDefaultMaterialAdapter::vertexColorGreenMask()
+{
+    return m_material.vertexColorGreenMask;
+}
+
+QSSGRenderDefaultMaterial::VertexColorMaskFlags QSSGShaderDefaultMaterialAdapter::vertexColorBlueMask()
+{
+    return m_material.vertexColorBlueMask;
+}
+
+QSSGRenderDefaultMaterial::VertexColorMaskFlags QSSGShaderDefaultMaterialAdapter::vertexColorAlphaMask()
+{
+    return m_material.vertexColorAlphaMask;
+}
+
 QSSGRenderImage *QSSGShaderDefaultMaterialAdapter::iblProbe()
 {
     return m_material.iblProbe;
@@ -150,11 +202,46 @@ float QSSGShaderDefaultMaterialAdapter::ior()
     return m_material.ior;
 }
 
+bool QSSGShaderDefaultMaterialAdapter::isFresnelScaleBiasEnabled()
+{
+    return m_material.fresnelScaleBiasEnabled;
+}
+
+float QSSGShaderDefaultMaterialAdapter::fresnelScale()
+{
+    return m_material.fresnelScale;
+}
+
+float QSSGShaderDefaultMaterialAdapter::fresnelBias()
+{
+    return m_material.fresnelBias;
+}
+
 float QSSGShaderDefaultMaterialAdapter::fresnelPower()
 {
     return m_material.fresnelPower;
 }
 
+bool QSSGShaderDefaultMaterialAdapter::isClearcoatFresnelScaleBiasEnabled()
+{
+    return m_material.clearcoatFresnelScaleBiasEnabled;
+}
+
+float QSSGShaderDefaultMaterialAdapter::clearcoatFresnelScale()
+{
+    return m_material.clearcoatFresnelScale;
+}
+
+float QSSGShaderDefaultMaterialAdapter::clearcoatFresnelBias()
+{
+    return m_material.clearcoatFresnelBias;
+}
+
+float QSSGShaderDefaultMaterialAdapter::clearcoatFresnelPower()
+{
+    return m_material.clearcoatFresnelPower;
+}
+
 float QSSGShaderDefaultMaterialAdapter::metalnessAmount()
 {
     return m_material.metalnessAmount;
@@ -230,6 +317,11 @@ float QSSGShaderDefaultMaterialAdapter::clearcoatRoughnessAmount()
     return m_material.clearcoatRoughnessAmount;
 }
 
+float QSSGShaderDefaultMaterialAdapter::clearcoatNormalStrength()
+{
+    return m_material.clearcoatNormalStrength;
+}
+
 float QSSGShaderDefaultMaterialAdapter::transmissionFactor()
 {
     return m_material.transmissionFactor;
@@ -284,6 +376,31 @@ bool QSSGShaderCustomMaterialAdapter::isVertexColorsEnabled()
     return m_material.m_renderFlags.testFlag(QSSGRenderCustomMaterial::RenderFlag::VarColor);
 }
 
+bool QSSGShaderCustomMaterialAdapter::isVertexColorsMaskEnabled()
+{
+    return false;
+}
+
+bool QSSGShaderCustomMaterialAdapter::isInvertOpacityMapValue()
+{
+    return false;
+}
+
+bool QSSGShaderCustomMaterialAdapter::isBaseColorSingleChannelEnabled()
+{
+    return false;
+}
+
+bool QSSGShaderCustomMaterialAdapter::isSpecularAmountSingleChannelEnabled()
+{
+    return false;
+}
+
+bool QSSGShaderCustomMaterialAdapter::isEmissiveSingleChannelEnabled()
+{
+    return false;
+}
+
 bool QSSGShaderCustomMaterialAdapter::isClearcoatEnabled()
 {
     // TODO: Expose Clearcoat properties to custom material
@@ -306,6 +423,11 @@ bool QSSGShaderCustomMaterialAdapter::usesCustomSkinning()
     return m_material.m_renderFlags.testFlag(QSSGRenderCustomMaterial::RenderFlag::Skinning);
 }
 
+bool QSSGShaderCustomMaterialAdapter::usesCustomMorphing()
+{
+    return m_material.m_renderFlags.testFlag(QSSGRenderCustomMaterial::RenderFlag::Morphing);
+}
+
 QSSGRenderDefaultMaterial::MaterialSpecularModel QSSGShaderCustomMaterialAdapter::specularModel()
 {
     return QSSGRenderDefaultMaterial::MaterialSpecularModel::Default;
@@ -316,6 +438,26 @@ QSSGRenderDefaultMaterial::MaterialAlphaMode QSSGShaderCustomMaterialAdapter::al
     return QSSGRenderDefaultMaterial::MaterialAlphaMode::Default;
 }
 
+QSSGRenderDefaultMaterial::VertexColorMaskFlags QSSGShaderCustomMaterialAdapter::vertexColorRedMask()
+{
+    return QSSGRenderDefaultMaterial::NoMask;
+}
+
+QSSGRenderDefaultMaterial::VertexColorMaskFlags QSSGShaderCustomMaterialAdapter::vertexColorGreenMask()
+{
+    return QSSGRenderDefaultMaterial::NoMask;
+}
+
+QSSGRenderDefaultMaterial::VertexColorMaskFlags QSSGShaderCustomMaterialAdapter::vertexColorBlueMask()
+{
+    return QSSGRenderDefaultMaterial::NoMask;
+}
+
+QSSGRenderDefaultMaterial::VertexColorMaskFlags QSSGShaderCustomMaterialAdapter::vertexColorAlphaMask()
+{
+    return QSSGRenderDefaultMaterial::NoMask;
+}
+
 QSSGRenderImage *QSSGShaderCustomMaterialAdapter::iblProbe()
 {
     return m_material.m_iblProbe;
@@ -348,11 +490,46 @@ float QSSGShaderCustomMaterialAdapter::ior()
     return 1.45f;
 }
 
+bool QSSGShaderCustomMaterialAdapter::isFresnelScaleBiasEnabled()
+{
+    return false;
+}
+
+float QSSGShaderCustomMaterialAdapter::fresnelScale()
+{
+    return 1.0f;
+}
+
+float QSSGShaderCustomMaterialAdapter::fresnelBias()
+{
+    return 0.0f;
+}
+
 float QSSGShaderCustomMaterialAdapter::fresnelPower()
 {
     return 0.0f;
 }
 
+bool QSSGShaderCustomMaterialAdapter::isClearcoatFresnelScaleBiasEnabled()
+{
+    return false;
+}
+
+float QSSGShaderCustomMaterialAdapter::clearcoatFresnelScale()
+{
+    return 1.0f;
+}
+
+float QSSGShaderCustomMaterialAdapter::clearcoatFresnelBias()
+{
+    return 0.0f;
+}
+
+float QSSGShaderCustomMaterialAdapter::clearcoatFresnelPower()
+{
+    return 0.0f;
+}
+
 float QSSGShaderCustomMaterialAdapter::metalnessAmount()
 {
     return 0.0f;
@@ -428,6 +605,11 @@ float QSSGShaderCustomMaterialAdapter::clearcoatRoughnessAmount()
     return 0.0f;
 }
 
+float QSSGShaderCustomMaterialAdapter::clearcoatNormalStrength()
+{
+    return 1.0f;
+}
+
 float QSSGShaderCustomMaterialAdapter::transmissionFactor()
 {
     return 0.0f;
@@ -462,26 +644,31 @@ bool QSSGShaderCustomMaterialAdapter::hasCustomShaderSnippet(QSSGShaderCache::Sh
 }
 
 QByteArray QSSGShaderCustomMaterialAdapter::customShaderSnippet(QSSGShaderCache::ShaderType type,
-                                                                const QSSGRef<QSSGShaderLibraryManager> &shaderLibraryManager)
+                                                                QSSGShaderLibraryManager &shaderLibraryManager,
+                                                                bool multiViewCompatible)
 {
-    if (hasCustomShaderSnippet(type))
-        return shaderLibraryManager->getShaderSource(m_material.m_shaderPathKey, type);
+    if (hasCustomShaderSnippet(type)) {
+        const QByteArray shaderPathKey = m_material.m_shaderPathKey[multiViewCompatible ? QSSGRenderCustomMaterial::MultiViewShaderPathKeyIndex
+                                                                                        : QSSGRenderCustomMaterial::RegularShaderPathKeyIndex];
+        return shaderLibraryManager.getShaderSource(shaderPathKey, type);
+    }
 
     return QByteArray();
 }
 
 bool QSSGShaderCustomMaterialAdapter::hasCustomShaderFunction(QSSGShaderCache::ShaderType shaderType,
                                                               const QByteArray &funcName,
-                                                              const QSSGRef<QSSGShaderLibraryManager> &shaderLibraryManager)
+                                                              QSSGShaderLibraryManager &shaderLibraryManager)
 {
     if (hasCustomShaderSnippet(shaderType))
-        return shaderLibraryManager->getShaderMetaData(m_material.m_shaderPathKey, shaderType).customFunctions.contains(funcName);
+        return shaderLibraryManager.getShaderMetaData(m_material.m_shaderPathKey[QSSGRenderCustomMaterial::RegularShaderPathKeyIndex],
+                                                      shaderType).customFunctions.contains(funcName);
 
     return false;
 }
 
 void QSSGShaderCustomMaterialAdapter::setCustomPropertyUniforms(char *ubufData,
-                                                                QSSGRef<QSSGRhiShaderPipeline> &shaderPipeline,
+                                                                QSSGRhiShaderPipeline &shaderPipeline,
                                                                 const QSSGRenderContextInterface &context)
 {
     context.customMaterialSystem()->applyRhiShaderPropertyValues(ubufData, m_material, shaderPipeline);
@@ -498,96 +685,84 @@ namespace {
 // Must be in sync with the shader generator.
 static const QSSGCustomMaterialVariableSubstitution qssg_var_subst_tab[] = {
     // uniform (block members)
-    { "MODELVIEWPROJECTION_MATRIX", "qt_modelViewProjection" },
-    { "VIEWPROJECTION_MATRIX", "qt_viewProjectionMatrix" },
-    { "MODEL_MATRIX", "qt_modelMatrix" },
-    { "VIEW_MATRIX", "qt_viewMatrix" },
-    { "NORMAL_MATRIX", "qt_normalMatrix"},
-    { "BONE_TRANSFORMS", "qt_getTexMatrix" },
-    { "BONE_NORMAL_TRANSFORMS", "qt_getTexMatrix" },
-    { "PROJECTION_MATRIX", "qt_projectionMatrix" },
-    { "INVERSE_PROJECTION_MATRIX", "qt_inverseProjectionMatrix" },
-    { "CAMERA_POSITION", "qt_cameraPosition" },
-    { "CAMERA_DIRECTION", "qt_cameraDirection" },
-    { "CAMERA_PROPERTIES", "qt_cameraProperties" },
-    { "FRAMEBUFFER_Y_UP", "qt_rhi_properties.x" },
-    { "NDC_Y_UP", "qt_rhi_properties.y" },
-    { "NEAR_CLIP_VALUE", "qt_rhi_properties.z" },
-    { "IBL_MAXMIPMAP", "qt_lightProbeProperties.y" },
-    { "IBL_HORIZON", "qt_lightProbeProperties.z" },
-    { "IBL_EXPOSE", "qt_lightProbeProperties.w" },
+    { "MODELVIEWPROJECTION_MATRIX", "qt_modelViewProjection", true },
+    { "VIEWPROJECTION_MATRIX", "qt_viewProjectionMatrix", true },
+    { "MODEL_MATRIX", "qt_modelMatrix", false },
+    { "VIEW_MATRIX", "qt_viewMatrix", true },
+    { "NORMAL_MATRIX", "qt_normalMatrix", false },
+    { "BONE_TRANSFORMS", "qt_getTexMatrix", false },
+    { "BONE_NORMAL_TRANSFORMS", "qt_getTexMatrix", false },
+    { "PROJECTION_MATRIX", "qt_projectionMatrix", true },
+    { "INVERSE_PROJECTION_MATRIX", "qt_inverseProjectionMatrix", true },
+    { "CAMERA_POSITION", "qt_cameraPosition", true },
+    { "CAMERA_DIRECTION", "qt_cameraDirection", true },
+    { "CAMERA_PROPERTIES", "qt_cameraProperties", false },
+    { "FRAMEBUFFER_Y_UP", "qt_rhi_properties.x", false },
+    { "NDC_Y_UP", "qt_rhi_properties.y", false },
+    { "NEAR_CLIP_VALUE", "qt_rhi_properties.z", false },
+    { "IBL_MAXMIPMAP", "qt_lightProbeProperties.y", false },
+    { "IBL_HORIZON", "qt_lightProbeProperties.z", false },
+    { "IBL_EXPOSE", "qt_lightProbeProperties.w", false },
 
     // outputs
-    { "POSITION", "gl_Position" },
-    { "FRAGCOLOR", "fragOutput" },
-    { "POINT_SIZE", "gl_PointSize" },
+    { "POSITION", "gl_Position", false },
+    { "FRAGCOLOR", "fragOutput", false },
+    { "POINT_SIZE", "gl_PointSize", false },
 
     // fragment inputs
-    { "FRAGCOORD", "gl_FragCoord"},
+    { "FRAGCOORD", "gl_FragCoord", false },
 
     // functions
-    { "DIRECTIONAL_LIGHT", "qt_directionalLightProcessor" },
-    { "POINT_LIGHT", "qt_pointLightProcessor" },
-    { "SPOT_LIGHT", "qt_spotLightProcessor" },
-    { "AMBIENT_LIGHT", "qt_ambientLightProcessor" },
-    { "SPECULAR_LIGHT", "qt_specularLightProcessor" },
-    { "MAIN", "qt_customMain" },
-    { "POST_PROCESS", "qt_customPostProcessor" },
-    { "IBL_PROBE", "qt_iblProbeProcessor" },
+    { "DIRECTIONAL_LIGHT", "qt_directionalLightProcessor", false },
+    { "POINT_LIGHT", "qt_pointLightProcessor", false },
+    { "SPOT_LIGHT", "qt_spotLightProcessor", false },
+    { "AMBIENT_LIGHT", "qt_ambientLightProcessor", false },
+    { "SPECULAR_LIGHT", "qt_specularLightProcessor", false },
+    { "MAIN", "qt_customMain", false },
+    { "POST_PROCESS", "qt_customPostProcessor", false },
+    { "IBL_PROBE", "qt_iblProbeProcessor", false },
 
     // textures
-    { "SCREEN_TEXTURE", "qt_screenTexture" },
-    { "SCREEN_MIP_TEXTURE", "qt_screenTexture" }, // same resource as SCREEN_TEXTURE under the hood
-    { "DEPTH_TEXTURE", "qt_depthTexture" },
-    { "AO_TEXTURE", "qt_aoTexture" },
-    { "IBL_TEXTURE", "qt_lightProbe" },
-    { "LIGHTMAP", "qt_lightmap" },
+    { "SCREEN_TEXTURE", "qt_screenTexture", false },
+    { "SCREEN_MIP_TEXTURE", "qt_screenTexture", false }, // same resource as SCREEN_TEXTURE under the hood
+    { "DEPTH_TEXTURE", "qt_depthTexture", false },
+    { "AO_TEXTURE", "qt_aoTexture", false },
+    { "IBL_TEXTURE", "qt_lightProbe", false },
+    { "LIGHTMAP", "qt_lightmap", false },
 
     // For shaded only: vertex outputs, for convenience and perf. (only those
     // that are always present when lighting is enabled) The custom vertex main
     // can also calculate on its own and pass them on with VARYING but that's a
     // bit wasteful since we calculate these anyways.
-    { "VAR_WORLD_NORMAL", "qt_varNormal" },
-    { "VAR_WORLD_TANGENT", "qt_varTangent" },
-    { "VAR_WORLD_BINORMAL", "qt_varBinormal" },
-    { "VAR_WORLD_POSITION", "qt_varWorldPos" },
+    { "VAR_WORLD_NORMAL", "qt_varNormal", false },
+    { "VAR_WORLD_TANGENT", "qt_varTangent", false },
+    { "VAR_WORLD_BINORMAL", "qt_varBinormal", false },
+    { "VAR_WORLD_POSITION", "qt_varWorldPos", false },
     // vertex color is always enabled for custom materials (shaded)
-    { "VAR_COLOR", "qt_varColor" },
+    { "VAR_COLOR", "qt_varColor", false },
 
     // effects
-    { "INPUT", "qt_inputTexture" },
-    { "INPUT_UV", "qt_inputUV" },
-    { "TEXTURE_UV", "qt_textureUV" },
-    { "INPUT_SIZE", "qt_inputSize" },
-    { "OUTPUT_SIZE", "qt_outputSize" },
-    { "FRAME", "qt_frame_num" },
+    { "INPUT", "qt_inputTexture", false },
+    { "INPUT_UV", "qt_inputUV", false },
+    { "TEXTURE_UV", "qt_textureUV", false },
+    { "INPUT_SIZE", "qt_inputSize", false },
+    { "OUTPUT_SIZE", "qt_outputSize", false },
+    { "FRAME", "qt_frame_num", false },
 
     // instancing
-    { "INSTANCE_COLOR", "qt_instanceColor"},
-    { "INSTANCE_DATA", "qt_instanceData"},
-    { "INSTANCE_INDEX", "gl_InstanceIndex"},
+    { "INSTANCE_COLOR", "qt_instanceColor", false },
+    { "INSTANCE_DATA", "qt_instanceData", false },
+    { "INSTANCE_INDEX", "gl_InstanceIndex", false },
 
     // morphing
-    { "MORPH_POSITION0", "attr_tpos0"},
-    { "MORPH_POSITION1", "attr_tpos1"},
-    { "MORPH_POSITION2", "attr_tpos2"},
-    { "MORPH_POSITION3", "attr_tpos3"},
-    { "MORPH_POSITION4", "attr_tpos4"},
-    { "MORPH_POSITION5", "attr_tpos5"},
-    { "MORPH_POSITION6", "attr_tpos6"},
-    { "MORPH_POSITION7", "attr_tpos7"},
-    { "MORPH_NORMAL0", "attr_tnorm0"},
-    { "MORPH_NORMAL1", "attr_tnorm1"},
-    { "MORPH_NORMAL2", "attr_tnorm2"},
-    { "MORPH_NORMAL3", "attr_tnorm3"},
-    { "MORPH_TANGENT0", "attr_ttan0"},
-    { "MORPH_TANGENT1", "attr_ttan1"},
-    { "MORPH_BINORMAL0", "attr_tbinorm0"},
-    { "MORPH_BINORMAL1", "attr_tbinorm1"},
-    { "MORPH_WEIGHTS", "qt_morphWeights"},
+    { "MORPH_POSITION", "qt_getTargetPositionFromTargetId", false },
+    { "MORPH_NORMAL", "qt_getTargetNormalFromTargetId", false },
+    { "MORPH_TANGENT", "qt_getTargetTangentFromTargetId", false },
+    { "MORPH_BINORMAL", "qt_getTargetBinormalFromTargetId", false },
+    { "MORPH_WEIGHTS", "qt_morphWeights", false },
 
     // custom variables
-    { "SHARED_VARS", "struct QT_SHARED_VARS"}
+    { "SHARED_VARS", "struct QT_SHARED_VARS", false }
 };
 
 // Functions that, if present, get an argument list injected.
@@ -712,7 +887,8 @@ QSSGShaderCustomMaterialAdapter::prepareCustomShader(QByteArray &dst,
                                                      QSSGShaderCache::ShaderType type,
                                                      const StringPairList &baseUniforms,
                                                      const StringPairList &baseInputs,
-                                                     const StringPairList &baseOutputs)
+                                                     const StringPairList &baseOutputs,
+                                                     bool multiViewCompatible)
 {
     QByteArrayList inputs;
     QByteArrayList outputs;
@@ -730,7 +906,7 @@ QSSGShaderCustomMaterialAdapter::prepareCustomShader(QByteArray &dst,
     //
     // NOTE: This is not perfect, we do expend the custom material and effect shaders, so
     // there cane still be cases where the reported line numbers are slightly off.
-    if (!QSSGRhiContext::shaderDebuggingEnabled())
+    if (!QSSGRhiContextPrivate::shaderDebuggingEnabled())
         result.prepend("#line 1\n");
     const char *lastPos = shaderCode.constData();
 
@@ -809,7 +985,11 @@ QSSGShaderCustomMaterialAdapter::prepareCustomShader(QByteArray &dst,
 
                 for (const QSSGCustomMaterialVariableSubstitution &subst : qssg_var_subst_tab) {
                     if (trimmedId == subst.builtin) {
-                        id.replace(subst.builtin, subst.actualName); // replace, not assignment, to keep whitespace etc.
+                        QByteArray newExpr;
+                        newExpr.assign(subst.actualName);
+                        if (subst.multiViewDependent && multiViewCompatible)
+                            newExpr += QByteArrayLiteral("[qt_viewIndex]");
+                        id.replace(subst.builtin, newExpr); // replace, not assignment, to keep whitespace etc.
                         if (trimmedId == QByteArrayLiteral("BONE_TRANSFORMS")) {
                             useJointTexState = 0;
                             md.flags |= QSSGCustomShaderMetaData::UsesSkinning;
@@ -817,7 +997,11 @@ QSSGShaderCustomMaterialAdapter::prepareCustomShader(QByteArray &dst,
                             useJointNormalTexState = 0;
                             md.flags |= QSSGCustomShaderMetaData::UsesSkinning;
                         }
-
+                        if (trimmedId == QByteArrayLiteral("MORPH_POSITION") ||
+                                trimmedId == QByteArrayLiteral("MORPH_NORMAL") ||
+                                trimmedId == QByteArrayLiteral("MORPH_TANGENT") ||
+                                trimmedId == QByteArrayLiteral("MORPH_BINORMAL"))
+                            md.flags |= QSSGCustomShaderMetaData::UsesMorphing;
                         break;
                     }
                 }
