@@ -44,8 +44,8 @@ void ShadowMapPass::renderPrep(QSSGRenderer &renderer, QSSGLayerRenderData &data
     Q_UNUSED(renderer)
     using namespace RenderHelpers;
 
-    camera = data.camera;
-    QSSG_ASSERT(camera, return);
+    QSSG_ASSERT(!data.renderedCameras.isEmpty(), return);
+    camera = data.renderedCameras[0];
 
     const auto &renderedDepthWriteObjects = data.getSortedRenderedDepthWriteObjects(*camera);
     const auto &renderedOpaqueDepthPrepassObjects = data.getSortedrenderedOpaqueDepthPrepassObjects(*camera);
@@ -136,8 +136,8 @@ void ReflectionMapPass::renderPrep(QSSGRenderer &renderer, QSSGLayerRenderData &
     Q_UNUSED(renderer);
     Q_UNUSED(data);
 
-    auto *camera = data.camera;
-    QSSG_ASSERT(camera, return);
+    QSSG_ASSERT(!data.renderedCameras.isEmpty(), return);
+    QSSGRenderCamera *camera = data.renderedCameras[0];
 
     ps = data.getPipelineState();
     ps.flags |= { QSSGRhiGraphicsPipelineState::Flag::DepthTestEnabled,
@@ -228,8 +228,8 @@ void ZPrePassPass::renderPrep(QSSGRenderer &renderer, QSSGLayerRenderData &data)
 
     // CONDITION: Input + globally enabled or ?
 
-    auto *camera = data.camera;
-    QSSG_ASSERT(camera, return);
+    QSSG_ASSERT(!data.renderedCameras.isEmpty(), return);
+    QSSGRenderCamera *camera = data.renderedCameras[0];
 
     const auto &rhiCtx = renderer.contextInterface()->rhiContext();
     QSSG_ASSERT(rhiCtx->rhi()->isRecordingFrame(), return);
@@ -292,8 +292,9 @@ void SSAOMapPass::renderPrep(QSSGRenderer &renderer, QSSGLayerRenderData &data)
 
     rhiAoTexture = data.getRenderResult(QSSGFrameData::RenderResult::AoTexture);
     rhiDepthTexture = data.getRenderResult(QSSGFrameData::RenderResult::DepthTexture);
-    camera = data.camera;
-    QSSG_ASSERT_X((camera && rhiDepthTexture && rhiDepthTexture->isValid()), "Preparing AO pass failed, missing camera or required texture(s)", return);
+    QSSG_ASSERT_X(!data.renderedCameras.isEmpty(), "Preparing AO pass failed, missing camera", return);
+    camera = data.renderedCameras[0];
+    QSSG_ASSERT_X((rhiDepthTexture && rhiDepthTexture->isValid()), "Preparing AO pass failed, missing equired texture(s)", return);
 
     const auto &shaderCache = renderer.contextInterface()->shaderCache();
     ssaoShaderPipeline = shaderCache->getBuiltInRhiShaders().getRhiSsaoShader();
@@ -360,8 +361,8 @@ void DepthMapPass::renderPrep(QSSGRenderer &renderer, QSSGLayerRenderData &data)
 {
     using namespace RenderHelpers;
 
-    auto *camera = data.camera;
-    QSSG_ASSERT(camera, return);
+    QSSG_ASSERT(!data.renderedCameras.isEmpty(), return);
+    QSSGRenderCamera *camera = data.renderedCameras[0];
 
     const auto &rhiCtx = renderer.contextInterface()->rhiContext();
     QSSG_ASSERT(rhiCtx->rhi()->isRecordingFrame(), return);
@@ -441,8 +442,8 @@ void ScreenMapPass::renderPrep(QSSGRenderer &renderer, QSSGLayerRenderData &data
 {
     using namespace RenderHelpers;
 
-    auto *camera = data.camera;
-    QSSG_ASSERT(camera, return);
+    QSSG_ASSERT(!data.renderedCameras.isEmpty(), return);
+    QSSGRenderCamera *camera = data.renderedCameras[0];
 
     const auto &rhiCtx = renderer.contextInterface()->rhiContext();
     QSSG_ASSERT(rhiCtx->rhi()->isRecordingFrame(), return);
@@ -482,7 +483,7 @@ void ScreenMapPass::renderPrep(QSSGRenderer &renderer, QSSGLayerRenderData &data
             // Reflection cube maps are not available at this point, make sure they are turned off.
             bool recRef = handle.obj->renderableFlags.receivesReflections();
             handle.obj->renderableFlags.setReceivesReflections(false);
-            rhiPrepareRenderable(rhiCtx.get(), this, data, *handle.obj, rhiScreenTexture->rpDesc, &ps, shaderFeatures, 1);
+            rhiPrepareRenderable(rhiCtx.get(), this, data, *handle.obj, rhiScreenTexture->rpDesc, &ps, shaderFeatures, 1, rhiCtx->mainPassViewCount());
             handle.obj->renderableFlags.setReceivesReflections(recRef);
         }
     }
@@ -555,8 +556,8 @@ void ScreenMapPass::resetForFrame()
 
 void ScreenReflectionPass::renderPrep(QSSGRenderer &renderer, QSSGLayerRenderData &data)
 {
-    auto *camera = data.camera;
-    QSSG_ASSERT(camera, return);
+    QSSG_ASSERT(!data.renderedCameras.isEmpty(), return);
+    QSSGRenderCamera *camera = data.renderedCameras[0];
 
     const auto &rhiCtx = renderer.contextInterface()->rhiContext();
     QSSG_ASSERT(rhiCtx->rhi()->isRecordingFrame(), return);
@@ -569,6 +570,7 @@ void ScreenReflectionPass::renderPrep(QSSGRenderer &renderer, QSSGLayerRenderDat
 
     QRhiRenderPassDescriptor *mainRpDesc = rhiCtx->mainRenderPassDescriptor();
     const int samples = rhiCtx->mainPassSampleCount();
+    const int viewCount = rhiCtx->mainPassViewCount();
 
     // NOTE: We're piggybacking on the screen map pass for now, but we could do better.
     ps = data.getPipelineState();
@@ -584,7 +586,7 @@ void ScreenReflectionPass::renderPrep(QSSGRenderer &renderer, QSSGLayerRenderDat
         const bool curDepthWriteEnabled = !(depthWriteMode == QSSGDepthDrawMode::Never || depthWriteMode == QSSGDepthDrawMode::OpaquePrePass
                                      || data.isZPrePassActive() || !layerEnableDepthTest);
         ps.flags.setFlag(QSSGRhiGraphicsPipelineState::Flag::DepthWriteEnabled, curDepthWriteEnabled);
-        RenderHelpers::rhiPrepareRenderable(rhiCtx.get(), this, data, *theObject, mainRpDesc, &ps, shaderFeatures, samples);
+        RenderHelpers::rhiPrepareRenderable(rhiCtx.get(), this, data, *theObject, mainRpDesc, &ps, shaderFeatures, samples, viewCount);
     }
 }
 
@@ -638,7 +640,7 @@ void OpaquePass::prep(const QSSGRenderContextInterface &ctx,
                                             depthWriteMode == QSSGDepthDrawMode::OpaquePrePass ||
                                             data.isZPrePassActive() || !layerEnableDepthTest);
         ps.flags.setFlag(QSSGRhiGraphicsPipelineState::Flag::DepthWriteEnabled, curDepthWriteEnabled);
-        RenderHelpers::rhiPrepareRenderable(rhiCtx.get(), passKey, data, *theObject, rpDesc, &ps, shaderFeatures, ps.samples);
+        RenderHelpers::rhiPrepareRenderable(rhiCtx.get(), passKey, data, *theObject, rpDesc, &ps, shaderFeatures, ps.samples, ps.viewCount);
     }
 }
 
@@ -660,12 +662,12 @@ void OpaquePass::renderPrep(QSSGRenderer &renderer, QSSGLayerRenderData &data)
     auto *ctx = renderer.contextInterface();
     const auto &rhiCtx = ctx->rhiContext();
     QSSG_ASSERT(rhiCtx->rhi()->isRecordingFrame(), return);
-    auto camera = data.camera;
-    const auto &cameraData = data.cameraData;
-    QSSG_ASSERT(camera && cameraData.has_value() , return);
+    QSSG_ASSERT(!data.renderedCameras.isEmpty() && data.renderedCameraData.has_value() , return);
+    QSSGRenderCamera *camera = data.renderedCameras[0];
 
     ps = data.getPipelineState();
     ps.samples = rhiCtx->mainPassSampleCount();
+    ps.viewCount = rhiCtx->mainPassViewCount();
     ps.depthFunc = QRhiGraphicsPipeline::LessOrEqual;
     ps.flags.setFlag(QSSGRhiGraphicsPipelineState::Flag::BlendEnabled, false);
 
@@ -718,7 +720,7 @@ void TransparentPass::prep(const QSSGRenderContextInterface &ctx,
         const bool curDepthWriteEnabled = (depthWriteMode == QSSGDepthDrawMode::Always && !zPrePassActive);
         ps.flags.setFlag(QSSGRhiGraphicsPipelineState::Flag::DepthWriteEnabled, curDepthWriteEnabled);
         if (!(theObject->renderableFlags.isCompletelyTransparent()))
-            RenderHelpers::rhiPrepareRenderable(rhiCtx.get(), passKey, data, *theObject, rpDesc, &ps, shaderFeatures, ps.samples);
+            RenderHelpers::rhiPrepareRenderable(rhiCtx.get(), passKey, data, *theObject, rpDesc, &ps, shaderFeatures, ps.samples, ps.viewCount);
     }
 }
 
@@ -743,14 +745,15 @@ void TransparentPass::renderPrep(QSSGRenderer &renderer, QSSGLayerRenderData &da
 {
     auto *ctx = renderer.contextInterface();
     const auto &rhiCtx = ctx->rhiContext();
-    auto camera = data.camera;
-    const auto &cameraData = data.cameraData;
-    QSSG_ASSERT(camera && cameraData.has_value(), return);
+
+    QSSG_ASSERT(!data.renderedCameras.isEmpty() && data.renderedCameraData.has_value() , return);
+    QSSGRenderCamera *camera = data.renderedCameras[0];
 
     QRhiRenderPassDescriptor *mainRpDesc = rhiCtx->mainRenderPassDescriptor();
 
     ps = data.getPipelineState();
     ps.samples = rhiCtx->mainPassSampleCount();
+    ps.viewCount = rhiCtx->mainPassViewCount();
 
     // transparent objects (or, without LayerEnableDepthTest, all objects)
     ps.flags.setFlag(QSSGRhiGraphicsPipelineState::Flag::BlendEnabled, true);
@@ -790,8 +793,8 @@ void SkyboxPass::renderPrep(QSSGRenderer &renderer, QSSGLayerRenderData &data)
     if (!skipPrep) {
         const auto &rhiCtx = renderer.contextInterface()->rhiContext();
         QSSG_ASSERT(rhiCtx->rhi()->isRecordingFrame(), return);
-        auto camera = data.camera;
-        QSSG_ASSERT(camera, return);
+        QSSG_ASSERT(!data.renderedCameras.isEmpty(), return);
+        QSSGRenderCamera *camera = data.renderedCameras[0];
         layer = &data.layer;
         QSSG_ASSERT(layer, return);
 
@@ -829,6 +832,7 @@ void SkyboxPass::renderPass(QSSGRenderer &renderer)
     QSSGRhiGraphicsPipelineStatePrivate::setShaderPipeline(ps, shaderPipeline.get());
     QRhiRenderPassDescriptor *rpDesc = rhiCtx->mainRenderPassDescriptor();
     ps.samples = rhiCtx->mainPassSampleCount();
+    ps.viewCount = rhiCtx->mainPassViewCount();
     renderer.rhiQuadRenderer()->recordRenderQuad(rhiCtx.get(), &ps, srb, rpDesc, { QSSGRhiQuadRenderer::DepthTest | QSSGRhiQuadRenderer::RenderBehind });
     Q_QUICK3D_PROFILE_END_WITH_STRING(QQuick3DProfiler::Quick3DRenderPass, 0, QByteArrayLiteral("skybox_map"));
 }
@@ -844,8 +848,8 @@ void SkyboxCubeMapPass::renderPrep(QSSGRenderer &renderer, QSSGLayerRenderData &
 {
     const auto &rhiCtx = renderer.contextInterface()->rhiContext();
     QSSG_ASSERT(rhiCtx->rhi()->isRecordingFrame(), return);
-    auto camera = data.camera;
-    QSSG_ASSERT(camera, return);
+    QSSG_ASSERT(!data.renderedCameras.isEmpty(), return);
+    QSSGRenderCamera *camera = data.renderedCameras[0];
     layer = &data.layer;
     QSSG_ASSERT(layer, return);
 
@@ -909,18 +913,24 @@ void Item2DPass::renderPrep(QSSGRenderer &renderer, QSSGLayerRenderData &data)
 
         auto layerPrepResult = data.layerPrepResult;
 
-        const auto &renderTarget = rhiCtx->renderTarget();
+        QRhiRenderTarget *renderTarget = rhiCtx->renderTarget();
         item2D->m_renderer->setDevicePixelRatio(renderTarget->devicePixelRatio());
         const QRect deviceRect(QPoint(0, 0), renderTarget->pixelSize());
+        const int viewCount = rhiCtx->mainPassViewCount();
+        QSSG_ASSERT(item2D->mvps.count() == viewCount, return);
         if (layer.scissorRect.isValid()) {
             QRect effScissor = layer.scissorRect & layerPrepResult.viewport.toRect();
             QMatrix4x4 correctionMat = correctMVPForScissor(layerPrepResult.viewport,
                                                             effScissor,
                                                             rhiCtx->rhi()->isYUpInNDC());
-            item2D->m_renderer->setProjectionMatrix(correctionMat * item2D->MVP);
+            for (int viewIndex = 0; viewIndex < viewCount; ++viewIndex) {
+                const QMatrix4x4 projectionMatrix = correctionMat * item2D->mvps[viewIndex];
+                item2D->m_renderer->setProjectionMatrix(projectionMatrix, viewIndex);
+            }
             item2D->m_renderer->setViewportRect(effScissor);
         } else {
-            item2D->m_renderer->setProjectionMatrix(item2D->MVP);
+            for (int viewIndex = 0; viewIndex < viewCount; ++viewIndex)
+                item2D->m_renderer->setProjectionMatrix(item2D->mvps[viewIndex], viewIndex);
             item2D->m_renderer->setViewportRect(RenderHelpers::correctViewportCoordinates(layerPrepResult.viewport, deviceRect));
         }
         item2D->m_renderer->setDeviceRect(deviceRect);
@@ -938,7 +948,9 @@ void Item2DPass::renderPrep(QSSGRenderer &renderer, QSSGLayerRenderData &data)
             item2D->m_rp = rhiCtx->mainRenderPassDescriptor()->newCompatibleRenderPassDescriptor();
             QSSG_CHECK(item2D->m_rp);
         }
-        item2D->m_renderer->setRenderTarget({ renderTarget, item2D->m_rp, rhiCtx->commandBuffer() });
+        QSGRenderTarget sgRt(renderTarget, item2D->m_rp, rhiCtx->commandBuffer());
+        sgRt.multiViewCount = rhiCtx->mainPassViewCount();
+        item2D->m_renderer->setRenderTarget(sgRt);
         delete oldRp;
         item2D->m_renderer->prepareSceneInline();
     }
@@ -974,8 +986,8 @@ void InfiniteGridPass::renderPrep(QSSGRenderer &renderer, QSSGLayerRenderData &d
 {
     const auto &rhiCtx = renderer.contextInterface()->rhiContext();
     QSSG_ASSERT(rhiCtx->rhi()->isRecordingFrame(), return);
-    auto camera = data.camera;
-    QSSG_ASSERT(camera, return);
+    QSSG_ASSERT(!data.renderedCameras.isEmpty(), return);
+    QSSGRenderCamera *camera = data.renderedCameras[0];
     layer = &data.layer;
     QSSG_ASSERT(layer, return);
 
@@ -1014,8 +1026,8 @@ void DebugDrawPass::renderPrep(QSSGRenderer &renderer, QSSGLayerRenderData &data
     const auto &rhiCtx = renderer.contextInterface()->rhiContext();
     QSSG_ASSERT(rhiCtx->rhi()->isRecordingFrame(), return);
     QSSGRhiContextPrivate *rhiCtxD = QSSGRhiContextPrivate::get(rhiCtx.get());
-    auto camera = data.camera;
-    QSSG_ASSERT(camera, return);
+    QSSG_ASSERT(!data.renderedCameras.isEmpty(), return);
+    QSSGRenderCamera *camera = data.renderedCameras[0];
 
     const auto &shaderCache = renderer.contextInterface()->shaderCache();
     debugObjectShader = shaderCache->getBuiltInRhiShaders().getRhiDebugObjectShader();
