@@ -6,7 +6,7 @@
 
 #include <QtQuick3DRuntimeRender/private/qssgrenderer_p.h>
 #include <QtQuick3DRuntimeRender/private/qssgrenderlight_p.h>
-#include <QtQuick3DRuntimeRender/private/qssgrendercontextcore_p.h>
+#include "../qssgrendercontextcore.h"
 #include <QtQuick3DRuntimeRender/private/qssgrendershadercache_p.h>
 #include <QtQuick3DRuntimeRender/private/qssgrendershaderlibrarymanager_p.h>
 #include <QtQuick3DRuntimeRender/private/qssgrendershadercodegenerator_p.h>
@@ -15,10 +15,10 @@
 
 QT_BEGIN_NAMESPACE
 
-QSSGMaterialVertexPipeline::QSSGMaterialVertexPipeline(const QSSGRef<QSSGProgramGenerator> &programGen,
+QSSGMaterialVertexPipeline::QSSGMaterialVertexPipeline(QSSGProgramGenerator &programGen,
                                                        const QSSGShaderDefaultMaterialKeyProperties &materialProperties,
                                                        QSSGShaderMaterialAdapter *materialAdapter)
-    : m_programGenerator(programGen)
+    : m_programGenerator(&programGen)
     , defaultMaterialShaderKeyProperties(materialProperties)
     , materialAdapter(materialAdapter)
     , hasCustomShadedMain(false)
@@ -100,7 +100,7 @@ static inline const char *customMainCallWithArguments(bool usesInstancing)
 
 void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMaterialKey &inKey,
                                                        const QSSGShaderFeatures &inFeatureSet,
-                                                       const QSSGRef<QSSGShaderLibraryManager> &shaderLibraryManager)
+                                                       QSSGShaderLibraryManager &shaderLibraryManager)
 {
     QSSGShaderGeneratorStageFlags theStages(QSSGProgramGenerator::defaultFlags());
     programGenerator()->beginProgram(theStages);
@@ -131,8 +131,10 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
     const bool blendParticles = defaultMaterialShaderKeyProperties.m_blendParticles.getValue(inKey);
     usesInstancing = defaultMaterialShaderKeyProperties.m_usesInstancing.getValue(inKey);
     m_hasSkinning = defaultMaterialShaderKeyProperties.m_boneCount.getValue(inKey) > 0;
-    const auto morphSize = defaultMaterialShaderKeyProperties.m_morphTargetCount.getValue(inKey);
+    const auto morphSize = defaultMaterialShaderKeyProperties.m_targetCount.getValue(inKey);
     m_hasMorphing = morphSize > 0;
+    m_viewCount = inFeatureSet.isSet(QSSGShaderFeatures::Feature::DisableMultiView)
+        ? 1 : defaultMaterialShaderKeyProperties.m_viewCount.getValue(inKey);
 
     vertexShader.addIncoming("attr_pos", "vec3");
     if (usesInstancing) {
@@ -161,14 +163,18 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
 
         vertexShader.addUniform("qt_boneTexture", "sampler2D");
     }
-    if (m_hasMorphing)
+    if (m_hasMorphing) {
+        vertexShader.addInclude("morphanim.glsllib");
         vertexShader.addUniformArray("qt_morphWeights", "float", morphSize);
+        vertexShader.addUniform("qt_morphTargetTexture", "sampler2DArray");
+    }
 
     const bool hasCustomVertexShader = materialAdapter->hasCustomShaderSnippet(QSSGShaderCache::ShaderType::Vertex);
     const bool hasCustomFragmentShader = materialAdapter->hasCustomShaderSnippet(QSSGShaderCache::ShaderType::Fragment);
     if (hasCustomVertexShader) {
         QByteArray snippet = materialAdapter->customShaderSnippet(QSSGShaderCache::ShaderType::Vertex,
-                                                                  shaderLibraryManager);
+                                                                  shaderLibraryManager,
+                                                                  m_viewCount >= 2);
         if (materialAdapter->hasCustomShaderFunction(QSSGShaderCache::ShaderType::Vertex,
                                                      QByteArrayLiteral("qt_customMain"),
                                                      shaderLibraryManager))
@@ -184,6 +190,14 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
                 m_hasSkinning = false;
             }
 
+            if (materialAdapter->usesCustomMorphing()) {
+                vertexShader.addInclude("morphanim_custom.glsllib");
+                if (morphSize > 0)
+                    vertexShader.addUniformArray("qt_morphWeights", "float", morphSize);
+                vertexShader.addUniform("qt_morphTargetTexture", "sampler2DArray");
+                m_hasMorphing = false;
+            }
+
             if (!materialAdapter->isUnshaded()) {
                 hasCustomShadedMain = true;
             }
@@ -214,12 +228,18 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
         vertexShader.append("    vec4 qt_vertColor = vec4(1.0);"); // must be 1,1,1,1 to not alter when multiplying with it
 
     if (!usesInstancing) {
-        vertexShader.addUniform("qt_modelViewProjection", "mat4");
+        if (m_viewCount < 2)
+            vertexShader.addUniform("qt_modelViewProjection", "mat4");
+        else
+            vertexShader.addUniformArray("qt_modelViewProjection", "mat4", m_viewCount);
     } else {
         // Must manualy calculate a MVP
         vertexShader.addUniform("qt_modelMatrix", "mat4");
         vertexShader.addUniform("qt_parentMatrix", "mat4");
-        vertexShader.addUniform("qt_viewProjectionMatrix", "mat4");
+        if (m_viewCount < 2)
+            vertexShader.addUniform("qt_viewProjectionMatrix", "mat4");
+        else
+            vertexShader.addUniformArray("qt_viewProjectionMatrix", "mat4", m_viewCount);
     }
 
     // The custom fragment main should be skipped if this is a
@@ -235,17 +255,33 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
         // condition we have to ensure the keywords (VIEW_MATRIX etc.) promised
         // by the documentation are available in *both* the custom vertex and
         // fragment shader snippets, even if only one of them is present.
-        vertexShader.addUniform("qt_viewProjectionMatrix", "mat4");
+        if (m_viewCount < 2) {
+            vertexShader.addUniform("qt_viewProjectionMatrix", "mat4");
+            vertexShader.addUniform("qt_viewMatrix", "mat4");
+            vertexShader.addUniform("qt_cameraPosition", "vec3");
+            vertexShader.addUniform("qt_cameraDirection", "vec3");
+            if (usesProjectionMatrix)
+                vertexShader.addUniform("qt_projectionMatrix", "mat4");
+            if (usesInvProjectionMatrix)
+                vertexShader.addUniform("qt_inverseProjectionMatrix", "mat4");
+        } else {
+            vertexShader.addUniformArray("qt_viewProjectionMatrix", "mat4", m_viewCount);
+            vertexShader.addUniformArray("qt_viewMatrix", "mat4", m_viewCount);
+            vertexShader.addUniformArray("qt_cameraPosition", "vec3", m_viewCount);
+            vertexShader.addUniformArray("qt_cameraDirection", "vec3", m_viewCount);
+            if (usesProjectionMatrix)
+                vertexShader.addUniformArray("qt_projectionMatrix", "mat4", m_viewCount);
+            if (usesInvProjectionMatrix)
+                vertexShader.addUniformArray("qt_inverseProjectionMatrix", "mat4", m_viewCount);
+        }
         vertexShader.addUniform("qt_modelMatrix", "mat4");
-        vertexShader.addUniform("qt_viewMatrix", "mat4");
         vertexShader.addUniform("qt_normalMatrix", "mat3");
-        vertexShader.addUniform("qt_cameraPosition", "vec3");
-        vertexShader.addUniform("qt_cameraDirection", "vec3");
         vertexShader.addUniform("qt_cameraProperties", "vec2");
-        if (usesProjectionMatrix)
-            vertexShader.addUniform("qt_projectionMatrix", "mat4");
-        if (usesInvProjectionMatrix)
-            vertexShader.addUniform("qt_inverseProjectionMatrix", "mat4");
+    }
+
+    if (m_viewCount >= 2) {
+        addFlatParameter("qt_viewIndex", "uint");
+        vertexShader.append("    qt_viewIndex = gl_ViewIndex;");
     }
 
     if (meshHasNormals) {
@@ -272,7 +308,7 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
         vertexShader.append("    qt_vertBinormal = attr_binormal;");
         vertexShader.addIncoming("attr_binormal", "vec3");
     }
-    if (meshHasColors && materialAdapter->isVertexColorsEnabled()) {
+    if (meshHasColors) {
         vertexShader.append("    qt_vertColor = attr_color;");
         vertexShader.addIncoming("attr_color", "vec4");
     }
@@ -297,7 +333,10 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
         else
             vertexShader.append("    mat4 qt_instancedModelMatrix =  qt_parentMatrix * transpose(qt_instanceMatrix) * qt_modelMatrix;");
         vertexShader.append("    mat3 qt_instancedNormalMatrix = mat3(transpose(inverse(qt_instancedModelMatrix)));");
-        vertexShader.append("    mat4 qt_instancedMVPMatrix = qt_viewProjectionMatrix * qt_instancedModelMatrix;");
+        if (m_viewCount < 2)
+            vertexShader.append("    mat4 qt_instancedMVPMatrix = qt_viewProjectionMatrix * qt_instancedModelMatrix;");
+        else
+            vertexShader.append("    mat4 qt_instancedMVPMatrix = qt_viewProjectionMatrix[qt_viewIndex] * qt_instancedModelMatrix;");
     }
 
     if (!materialAdapter->isUnshaded() || !hasCustomVertexShader) {
@@ -308,7 +347,7 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
             vertexShader.append(customMainCallWithArguments(usesInstancing));
 
         if (m_hasMorphing && !hasCustomVertexShader)
-            vertexShader.append("    qt_vertPosition.xyz = qt_getMorphPosition(qt_vertPosition.xyz);");
+            vertexShader.append("    qt_vertPosition.xyz = qt_getTargetPosition(qt_vertPosition.xyz);");
 
         if (m_hasSkinning) {
             vertexShader.append("    mat4 skinMat = mat4(1);");
@@ -322,10 +361,14 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
         }
 
         if (!hasCustomShadedMain || !overridesPosition) {
-            if (!usesInstancing)
-                vertexShader.append("    gl_Position = qt_modelViewProjection * qt_vertPosition;");
-            else
+            if (!usesInstancing) {
+                if (m_viewCount < 2)
+                    vertexShader.append("    gl_Position = qt_modelViewProjection * qt_vertPosition;");
+                else
+                    vertexShader.append("    gl_Position = qt_modelViewProjection[qt_viewIndex] * qt_vertPosition;");
+            } else {
                 vertexShader.append("    gl_Position = qt_instancedMVPMatrix * qt_vertPosition;");
+            }
         }
     }
 
@@ -335,14 +378,15 @@ void QSSGMaterialVertexPipeline::beginVertexGeneration(const QSSGShaderDefaultMa
     } // with a custom vertex shader it is up to it to set gl_PointSize (aka POINT_SIZE)
 }
 
-void QSSGMaterialVertexPipeline::beginFragmentGeneration(const QSSGRef<QSSGShaderLibraryManager> &shaderLibraryManager)
+void QSSGMaterialVertexPipeline::beginFragmentGeneration(QSSGShaderLibraryManager &shaderLibraryManager)
 {
     fragment().addUniform("qt_material_properties", "vec4");
     fragment().addUniform("qt_rhi_properties", "vec4");
 
     if (!skipCustomFragmentSnippet && materialAdapter->hasCustomShaderSnippet(QSSGShaderCache::ShaderType::Fragment)) {
         QByteArray snippet = materialAdapter->customShaderSnippet(QSSGShaderCache::ShaderType::Fragment,
-                                                                  shaderLibraryManager);
+                                                                  shaderLibraryManager,
+                                                                  m_viewCount >= 2);
         if (!materialAdapter->isUnshaded()) {
             const bool usesShared = materialAdapter->usesSharedVariables();
             insertAmbientLightProcessorArgs(snippet, usesShared);
@@ -378,7 +422,7 @@ void QSSGMaterialVertexPipeline::doGenerateWorldNormal(const QSSGShaderDefaultMa
     if (!usesInstancing)
         vertexGenerator.addUniform("qt_normalMatrix", "mat3");
     if (m_hasMorphing)
-        vertexGenerator.append("    qt_vertNormal = qt_getMorphNormal(qt_vertNormal);");
+        vertexGenerator.append("    qt_vertNormal = qt_getTargetNormal(qt_vertNormal);");
     if (m_hasSkinning) {
         vertexGenerator.append("    if (qt_vertWeights != vec4(0.0))");
         vertexGenerator.append("        qt_vertNormal = qt_getSkinNormalMatrix(qt_vertJoints, qt_vertWeights) * qt_vertNormal;");
@@ -399,7 +443,7 @@ void QSSGMaterialVertexPipeline::doGenerateWorldNormal(const QSSGShaderDefaultMa
 void QSSGMaterialVertexPipeline::doGenerateVarTangent(const QSSGShaderDefaultMaterialKey &inKey)
 {
     if (m_hasMorphing)
-        vertex() << "    qt_vertTangent = qt_getMorphTangent(qt_vertTangent);\n";
+        vertex() << "    qt_vertTangent = qt_getTargetTangent(qt_vertTangent);\n";
     if (m_hasSkinning) {
         vertex() << "    if (qt_vertWeights != vec4(0.0))\n"
                  << "       qt_vertTangent = (skinMat * vec4(qt_vertTangent, 0.0)).xyz;\n";
@@ -419,7 +463,7 @@ void QSSGMaterialVertexPipeline::doGenerateVarTangent(const QSSGShaderDefaultMat
 void QSSGMaterialVertexPipeline::doGenerateVarBinormal(const QSSGShaderDefaultMaterialKey &inKey)
 {
     if (m_hasMorphing)
-        vertex() << "    qt_vertBinormal = qt_getMorphBinormal(qt_vertBinormal);\n";
+        vertex() << "    qt_vertBinormal = qt_getTargetBinormal(qt_vertBinormal);\n";
     if (m_hasSkinning) {
         vertex() << "    if (qt_vertWeights != vec4(0.0))\n"
                  << "       qt_vertBinormal = (skinMat * vec4(qt_vertBinormal, 0.0)).xyz;\n";
@@ -458,11 +502,16 @@ void QSSGMaterialVertexPipeline::endFragmentGeneration()
 
 void QSSGMaterialVertexPipeline::addInterpolationParameter(const QByteArray &inName, const QByteArray &inType)
 {
-    m_interpolationParameters.insert(inName, inType);
     vertex().addOutgoing(inName, inType);
     fragment().addIncoming(inName, inType);
 }
 
+void QSSGMaterialVertexPipeline::addFlatParameter(const QByteArray &inName, const QByteArray &inType)
+{
+    vertex().addFlatOutgoing(inName, inType);
+    fragment().addFlatIncoming(inName, inType);
+}
+
 QSSGStageGeneratorBase &QSSGMaterialVertexPipeline::activeStage()
 {
     return vertex();
